/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "houdini/houdini.pmh"

#define FE_SAH_ATOM_DEBUG	FALSE

namespace fe
{

void SurfaceAccessibleHoudini::setGdp(const GU_Detail* a_pGdp)
{
	safeLock();

	m_primitiveMap.clear();

//	if(m_pGdp!=a_pGdp || m_pGdpChangeable)
	{
		m_pGdp=a_pGdp;
		m_pGdpChangeable=NULL;
		m_spSurfaceI=NULL;
	}

	if(m_spSurfaceI.isValid())
	{
		sp<SurfaceTrianglesHoudini> spSurface=m_spSurfaceI;
		if(spSurface.isValid())
		{
			spSurface->setGdp(a_pGdp);
		}
	}

	safeUnlock();
}

void SurfaceAccessibleHoudini::setGdpChangeable(GU_Detail* a_pGdp)
{
	safeLock();

	m_primitiveMap.clear();

//	if(m_pGdp!=a_pGdp || m_pGdpChangeable!=a_pGdp)
	{
		m_pGdp=a_pGdp;
		m_pGdpChangeable=a_pGdp;
		m_spSurfaceI=NULL;
	}

	if(m_spSurfaceI.isValid())
	{
		sp<SurfaceTrianglesHoudini> spSurface=m_spSurfaceI;
		if(spSurface.isValid())
		{
			spSurface->setGdp(a_pGdp);
		}
	}

	safeUnlock();
}

sp<SurfaceAccessorI> SurfaceAccessibleHoudini::accessor(
	SurfaceAccessibleI::Element a_element,
	String a_name,
	SurfaceAccessibleI::Creation a_create)
{
	sp<SurfaceAccessorHoudini> spAccessor=accessor();
	if(spAccessor.isValid() &&
			(spAccessor->bind(a_element,a_name) ||
			(m_pGdpChangeable &&
			a_create==SurfaceAccessibleI::e_createMissing)))
	{
		spAccessor->setSurfaceAccessible(
				sp<SurfaceAccessibleI>(this));
		return accessorOf(spAccessor);
	}
	return sp<SurfaceAccessorI>(NULL);
}

sp<SurfaceAccessorI> SurfaceAccessibleHoudini::accessor(
	SurfaceAccessibleI::Element a_element,
	SurfaceAccessibleI::Attribute a_attribute,
	SurfaceAccessibleI::Creation a_create)
{
	sp<SurfaceAccessorHoudini> spAccessor=accessor();
	if(spAccessor.isValid() &&
			(spAccessor->bind(a_element,a_attribute) ||
			(m_pGdpChangeable &&
			a_create==SurfaceAccessibleI::e_createMissing)))
	{
		spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));
		return accessorOf(spAccessor);
	}
	return sp<SurfaceAccessorI>(NULL);
}

sp<SurfaceAccessorHoudini> SurfaceAccessibleHoudini::accessor(void)
{
	if(!m_pGdp)
	{
		return sp<SurfaceAccessorI>(NULL);
	}
	sp<SurfaceAccessorHoudini> spAccessorHoudini(new SurfaceAccessorHoudini);
	if(m_pGdpChangeable)
	{
		spAccessorHoudini->setGdpChangeable(m_pGdpChangeable);
	}
	else if(m_pGdp)
	{
		spAccessorHoudini->setGdp(m_pGdp);
	}

	return spAccessorHoudini;
}
sp<SurfaceAccessorI> SurfaceAccessibleHoudini::accessorOf(
	sp<SurfaceAccessorI> a_spAccessor)
{
//#ifdef FE_HOUDINI_USE_GA
#if FALSE
	FEASSERT(a_spAccessor.isValid());
	if(!m_pGdpChangeable || threading()==SurfaceAccessibleI::e_singleThread)
	{
		return a_spAccessor;
	}

	sp<SurfaceAccessorCached> spAccessorCached=
			registry()->create("*.SurfaceAccessorCached");
	FEASSERT(spAccessorCached.isValid());
	spAccessorCached->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));
	spAccessorCached->setAccessorChain(a_spAccessor);
	return spAccessorCached;
#else
	return a_spAccessor;
#endif
}

BWORD SurfaceAccessibleHoudini::discard(SurfaceAccessibleI::Element a_element,
	String a_name)
{
	if(!m_pGdpChangeable)
	{
		return FALSE;
	}

#ifdef FE_HOUDINI_USE_GA
	switch(a_element)
	{
		case SurfaceAccessibleI::e_point:
			m_pGdpChangeable->destroyPointAttrib(a_name.c_str());
			break;
		case SurfaceAccessibleI::e_vertex:
			m_pGdpChangeable->destroyVertexAttrib(a_name.c_str());
			break;
		case SurfaceAccessibleI::e_primitive:
			m_pGdpChangeable->destroyPrimAttrib(a_name.c_str());
			break;
		case SurfaceAccessibleI::e_detail:
			m_pGdpChangeable->destroyGlobalAttrib(a_name.c_str());
			break;
		default:
			return FALSE;
	}
#else
	switch(a_element)
	{
		case SurfaceAccessibleI::e_point:
		{
			GEO_PointAttribDict& attrDict=
					m_pGdpChangeable->pointAttribs();
			GB_Attribute* pAttr=attrDict.find(a_name.c_str());
			if(pAttr)
			{
				GB_AttribType attrType=pAttr->getType();
				int attrSize=pAttr->getSize();
				m_pGdpChangeable->destroyPointAttrib(a_name.c_str(),
						attrSize,attrType);
			}
		}
			break;
		case SurfaceAccessibleI::e_vertex:
		{
			GEO_VertexAttribDict& attrDict=
					m_pGdpChangeable->vertexAttribs();
			GB_Attribute* pAttr=attrDict.find(a_name.c_str());
			if(pAttr)
			{
				GB_AttribType attrType=pAttr->getType();
				int attrSize=pAttr->getSize();
				m_pGdpChangeable->destroyVertexAttrib(a_name.c_str(),
						attrSize,attrType);
			}
		}
			break;
		case SurfaceAccessibleI::e_primitive:
		{
			GEO_PrimAttribDict& attrDict=
					m_pGdpChangeable->primitiveAttribs();
			GB_Attribute* pAttr=attrDict.find(a_name.c_str());
			if(pAttr)
			{
				GB_AttribType attrType=pAttr->getType();
				int attrSize=pAttr->getSize();
				m_pGdpChangeable->destroyPrimAttrib(a_name.c_str(),
						attrSize,attrType);
			}
		}
			break;
		case SurfaceAccessibleI::e_detail:
		{
			GB_AttributeTable& attrDict=
					m_pGdpChangeable->attribs();
			GB_Attribute* pAttr=attrDict.find(a_name.c_str());
			if(pAttr)
			{
				GB_AttribType attrType=pAttr->getType();
				int attrSize=pAttr->getSize();
				m_pGdpChangeable->destroyAttrib(a_name.c_str(),
						attrSize,attrType);
			}
		}
			break;
		default:
			return FALSE;
	}
#endif

	return TRUE;
}

BWORD SurfaceAccessibleHoudini::discard(SurfaceAccessibleI::Element a_element,
	SurfaceAccessibleI::Attribute a_attribute)
{
	String name;
	switch(a_attribute)
	{
		case SurfaceAccessibleI::e_position:
			name="P";
			break;
		case SurfaceAccessibleI::e_normal:
			name="N";
			break;
		default:
			return FALSE;
	}

	return discard(a_element,name);
}

void SurfaceAccessibleHoudini::append(
	sp<SurfaceAccessibleI> a_spSurfaceAccessibleI,
	const SpatialTransform& a_rTransform)
{
	safeLock();

//	feLog("SurfaceAccessibleHoudini::append\n");
	sp<SurfaceAccessibleHoudini> spSurfaceOther=
			a_spSurfaceAccessibleI;
	if(!spSurfaceOther.isValid())
	{
		safeUnlock();

		//* TODO support cross copies
		feLog("SurfaceAccessibleHoudini::append"
				" not SurfaceAccessibleHoudini\n");
		return;
	}

	UT_Matrix4 utMatrix4(
			a_rTransform(0,0),a_rTransform(1,0),a_rTransform(2,0),0.0,
			a_rTransform(0,1),a_rTransform(1,1),a_rTransform(2,1),0.0,
			a_rTransform(0,2),a_rTransform(1,2),a_rTransform(2,2),0.0,
			a_rTransform(0,3),a_rTransform(1,3),a_rTransform(2,3),1.0);

	const GU_Detail* pOtherGdp=spSurfaceOther->gdp();
	GU_Detail copyGdp;
	copyGdp.duplicate(*pOtherGdp);
	copyGdp.transform(utMatrix4);
	const_cast<GU_Detail*>(m_pGdp)->merge(copyGdp);

	m_primitiveMap.clear();

	safeUnlock();
}

void SurfaceAccessibleHoudini::instance(
	sp<SurfaceAccessibleI> a_spSurfaceAccessibleI,
	const std::vector<SpatialTransform>& a_rTransformArray)
{
	//* TODO
}

void SurfaceAccessibleHoudini::initPrimitiveMap(void)
{
	if(!m_pGdp)
	{
		return;
	}

	m_hasDiscretePolygons=FALSE;

#ifdef FE_HOUDINI_USE_GA
	const I32 primitiveCount=m_pGdp->getNumPrimitives();
#else
	const I32 primitiveCount=m_pGdp->primitives().entries();
#endif

	for(I32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
#ifdef FE_HOUDINI_USE_GA
		const GEO_Primitive* pPrimitive=m_pGdp->getGEOPrimitive(
				m_pGdp->primitiveOffset(primitiveIndex));
#else
		const GEO_Primitive* pPrimitive=m_pGdp->primitives()(primitiveIndex);
#endif

#ifdef FE_HOUDINI_USE_GA
		if(pPrimitive->getTypeId().get()==GEO_PRIMPOLY)
#else
		if(pPrimitive->getPrimitiveId()==GEOPRIMPOLY)
#endif
		{
			m_hasDiscretePolygons=TRUE;
			break;
		}
	}

	I32 mapCount=0;
	if(m_hasDiscretePolygons)
	{
		m_primitiveMap[mapCount++]= -1;
	}

	for(I32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
#ifdef FE_HOUDINI_USE_GA
		const GEO_Primitive* pPrimitive=m_pGdp->getGEOPrimitive(
				m_pGdp->primitiveOffset(primitiveIndex));
#else
		const GEO_Primitive* pPrimitive=m_pGdp->primitives()(primitiveIndex);
#endif

#ifdef FE_HOUDINI_USE_GA
		if(pPrimitive->getTypeId().get()==GEO_PRIMPOLY)
#else
		if(pPrimitive->getPrimitiveId()==GEOPRIMPOLY)
#endif
		{
			continue;
		}

		m_primitiveMap[mapCount++]=primitiveIndex;
	}
}

I32 SurfaceAccessibleHoudini::lookupPrimitive(I32 a_surfaceIndex) const
{
	std::map<I32,I32>::const_iterator it=m_primitiveMap.find(a_surfaceIndex);
	if(it==m_primitiveMap.end())
	{
		return -2;
	}
	return it->second;
}

//* TODO generalize using pointPage in SurfaceAccessibleBase
#ifdef FE_HOUDINI_USE_GA
sp<SurfaceRange> SurfaceAccessibleHoudini::atomize(AtomicChange a_atomicChange,
	String a_group,U32 a_desiredCount)
{
#if FE_SAH_ATOM_DEBUG
	feLog("SurfaceAccessibleHoudini::atomize\n");
#endif

	//* TODO use a_desiredCount

	if(a_atomicChange==e_pointsOfPrimitives ||
			a_atomicChange==e_primitivesWithPoints)
	{
//		return atomizeCrossover(a_atomicChange,a_group);
		return atomizeConnectivity(a_atomicChange,a_group,a_desiredCount);
	}

	sp<SurfaceRange> spSurfaceRange(new SurfaceRange());

	const BWORD asPoints=(a_atomicChange==e_pointsOnly);

	const GA_ElementGroup* pGroup=NULL;
	if(!a_group.empty())
	{
		if(asPoints)
		{
			pGroup=m_pGdp->findPointGroup(a_group.c_str());
		}
		else
		{
			pGroup=m_pGdp->findPrimitiveGroup(a_group.c_str());
		}
	}

	GA_Range range=asPoints?
			m_pGdp->getPointRange((const GA_PointGroup*)pGroup):
			m_pGdp->getPrimitiveRange((const GA_PrimitiveGroup*)pGroup);
	GA_SplittableRange splitRange(range);
	for(GA_PageIterator pit=splitRange.beginPages();!pit.atEnd();++pit)
	{
		SurfaceRange::MultiSpan& rMultiSpan=spSurfaceRange->addAtomic();

		GA_Offset start,end;
		for(GA_Iterator it(pit.begin());it.blockAdvance(start,end); )
		{
			const U32 startIndex=asPoints?
					m_pGdp->pointIndex(start): m_pGdp->primitiveIndex(start);
			const U32 endIndex=asPoints?
					m_pGdp->pointIndex(end-1): m_pGdp->primitiveIndex(end-1);

#if FE_SAH_ATOM_DEBUG
			feLog("createRange span offset %d %d index %d %d\n",
					start,end,startIndex,endIndex);
#endif

			rMultiSpan.add(startIndex,endIndex);
		}
	}

#if FE_SAH_ATOM_DEBUG
	feLog("brief: %s\n",spSurfaceRange->brief().c_str());
	feLog("dump:\n%s\n",spSurfaceRange->dump().c_str());
#endif

	return spSurfaceRange;
}

void SurfaceAccessibleHoudini::preparePaging(AtomicChange a_atomicChange,
	String a_group)
{
	BWORD asPoints=(a_atomicChange==e_pointsOnly);

	//* POINTS
	const U32 pointCount=m_pGdp->getNumPoints();

	m_pageOfPoint.clear();
	m_pageOfPoint.resize(pointCount,-1);

	const GA_ElementGroup* pGroup=NULL;
	if(asPoints && !a_group.empty())
	{
		pGroup=m_pGdp->findPointGroup(a_group.c_str());
	}

	U32 pageIndex=1;
	GA_Range pointRange=
			m_pGdp->getPointRange((const GA_PointGroup*)pGroup);
	GA_SplittableRange splitPointRange(pointRange);
	for(GA_PageIterator pit=splitPointRange.beginPages();!pit.atEnd();++pit)
	{
		GA_Offset start,end;
		for(GA_Iterator it(pit.begin());it.blockAdvance(start,end); )
		{
			const U32 startIndex=m_pGdp->pointIndex(start);
			const U32 endIndex=m_pGdp->pointIndex(end-1);

			for(U32 m=startIndex;m<=endIndex;m++)
			{
				m_pageOfPoint[m]=pageIndex;
			}
		}
		pageIndex++;
	}

	if(!asPoints)
	{
		//* PRIMITIVES
		const U32 primitiveCount=m_pGdp->getNumPrimitives();

		m_pageOfPrimitive.clear();
		m_pageOfPrimitive.resize(primitiveCount,-1);

		pGroup=NULL;
		if(!asPoints && !a_group.empty())
		{
			pGroup=m_pGdp->findPrimitiveGroup(a_group.c_str());
		}

		pageIndex=1;
		GA_Range primitiveRange=
				m_pGdp->getPrimitiveRange((const GA_PrimitiveGroup*)pGroup);
		GA_SplittableRange splitPrimitiveRange(primitiveRange);
		for(GA_PageIterator pit=splitPrimitiveRange.beginPages();
				!pit.atEnd();++pit)
		{
			GA_Offset start,end;
			for(GA_Iterator it(pit.begin());it.blockAdvance(start,end); )
			{
				const U32 startIndex=m_pGdp->primitiveIndex(start);
				const U32 endIndex=m_pGdp->primitiveIndex(end-1);

				for(U32 m=startIndex;m<=endIndex;m++)
				{
					m_pageOfPrimitive[m]=pageIndex;
				}
			}
			pageIndex++;
		}
	}
}

I32 SurfaceAccessibleHoudini::pointPage(U32 a_pointIndex) const
{
	FEASSERT(!m_pageOfPoint.size() ||
			a_pointIndex<m_pageOfPoint.size());

	return a_pointIndex<m_pageOfPoint.size()?
			m_pageOfPoint[a_pointIndex]: 0;
}

I32 SurfaceAccessibleHoudini::primitivePage(U32 a_primitiveIndex) const
{
	FEASSERT(!m_pageOfPrimitive.size() ||
			a_primitiveIndex<m_pageOfPrimitive.size());

	return a_primitiveIndex<m_pageOfPrimitive.size()?
			m_pageOfPrimitive[a_primitiveIndex]: 0;
}

#endif	// FE_HOUDINI_USE_GA

#if FALSE
//* TODO remove deprecated
sp<SurfaceRange> SurfaceAccessibleHoudini::atomizeCrossover(
		AtomicChange a_atomicChange,String a_group)
{
	sp<SurfaceRange> spSurfaceRange(new SurfaceRange());

	const GA_PrimitiveGroup* pGroup=a_group.empty()? NULL:
			m_pGdp->findPrimitiveGroup(a_group.c_str());

	const U32 pointCount=m_pGdp->getNumPoints();
	const U32 primitiveCount=
			pGroup? pGroup->entries(): m_pGdp->getNumPrimitives();

	if(a_atomicChange==e_primitivesWithPoints)
	{
		//* TEMP one big atom
		SurfaceRange::MultiSpan& rMultiSpan=spSurfaceRange->addAtomic();
		rMultiSpan.add(0,primitiveCount);
		return spSurfaceRange;
	}

	std::vector<U32> pointAtom;
	std::vector<U32> primitiveAtom;

	pointAtom.resize(pointCount);
	primitiveAtom.resize(primitiveCount);

	for(U32 m=0;m<pointCount;m++)
	{
		pointAtom[m]=0;
	}
	for(U32 m=0;m<primitiveCount;m++)
	{
		primitiveAtom[m]=0;
	}

	U32 maxPage=1;

	GA_Range pointRange=m_pGdp->getPointRange();
	GA_SplittableRange splitPointRange(pointRange);
	for(GA_PageIterator pit=splitPointRange.beginPages();!pit.atEnd();++pit)
	{
		GA_Offset start,end;
		for(GA_Iterator it(pit.begin());it.blockAdvance(start,end); )
		{
			const U32 startIndex=m_pGdp->pointIndex(start);
			const U32 endIndex=m_pGdp->pointIndex(end-1);

			for(U32 m=startIndex;m<=endIndex;m++)
			{
				pointAtom[m]=maxPage;
			}
		}
		maxPage++;
	}

	GA_Range primitiveRange=m_pGdp->getPrimitiveRange(pGroup);
	GA_SplittableRange splitPrimitiveRange(primitiveRange);
	for(GA_PageIterator pit=splitPrimitiveRange.beginPages();!pit.atEnd();++pit)
	{
		GA_Offset start,end;
		for(GA_Iterator it(pit.begin());it.blockAdvance(start,end); )
		{
			for(U32 m=start;m<end;m++)
			{
				const U32 primitiveIndex=m_pGdp->primitiveIndex(m);

				const GEO_Primitive* pPrimitive=m_pGdp->getGEOPrimitive(m);

				U32 firstPage=0;

				const U32 subCount=pPrimitive->getVertexCount();
				U32 subIndex;
				for(subIndex=0;subIndex<subCount;subIndex++)
				{
					const GA_Offset offset=
							pPrimitive->getPointOffset(subIndex);
					const U32 pointIndex=m_pGdp->pointIndex(offset);

					const U32 page=pointAtom[pointIndex];

					if(!firstPage)
					{
						firstPage=pointAtom[pointIndex];
					}
					else if(pointAtom[pointIndex]!=firstPage)
					{
						feLog("crossover prim %d point %d page %d vs %d\n",
								primitiveIndex,pointIndex,page,firstPage);

						primitiveAtom[primitiveIndex]=0;
						break;
					}
				}

				if(subIndex==subCount)
				{
					primitiveAtom[primitiveIndex]=firstPage;
				}
			}
		}
	}

	for(U32 page=0;page<maxPage;page++)
	{
		SurfaceRange::MultiSpan* pMultiSpan=NULL;
		BWORD spanning=FALSE;
		U32 start;
		for(U32 m=0;m<primitiveCount;m++)
		{
			if(primitiveAtom[m]==page)
			{
				if(!pMultiSpan)
				{
					pMultiSpan=page? &spSurfaceRange->addAtomic():
							&spSurfaceRange->postAtomic();
				}
				if(!spanning)
				{
					start=m;
					spanning=TRUE;
				}
			}
			else
			{
				if(spanning)
				{
					pMultiSpan->add(start,m-1);
					spanning=FALSE;
				}
			}
		}
		if(spanning)
		{
			pMultiSpan->add(start,primitiveCount-1);
		}
	}

	feLog("brief: %s\n",spSurfaceRange->brief().c_str());
	feLog("dump:\n%s\n",spSurfaceRange->dump().c_str());

	return spSurfaceRange;
}
#endif

//* TODO SurfaceBag
sp<SurfaceI> SurfaceAccessibleHoudini::surface(String a_group)
{
	if(!m_pGdp)
	{
		return sp<SurfaceI>(NULL);
	}

	safeLock();

	if(m_spSurfaceI.isNull())
	{
#ifdef FE_HOUDINI_USE_GA
		const I32 entries=m_pGdp->getNumPrimitives();
#else
		const I32 entries=m_pGdp->primitives().entries();
#endif

		if(!entries)
		{
			safeUnlock();

			return sp<SurfaceI>(NULL);
		}

		if(entries==1)
		{
			safeUnlock();

			//* single primitive (perhaps a sphere)
			return subSurface(0,a_group);
		}

#ifdef FE_HOUDINI_USE_GA
		const GEO_Primitive* pPrimitive=
				m_pGdp->getGEOPrimitive(m_pGdp->primitiveOffset(0));
#else
		const GEO_Primitive* pPrimitive=m_pGdp->primitives()(0);
#endif

		//* HACK presuming they're all the same
		const GEO_Face* pGeoFace=dynamic_cast<const GEO_Face*>(pPrimitive);
		const BWORD openPoly=pGeoFace && !pGeoFace->isClosed();
		if(openPoly)
		{
			//* HACK presuming uniform CV counts
			const U32 subCount=pGeoFace->getVertexCount();

			//* all curves
			m_spSurfaceI=registry()->create("SurfaceI.SurfaceCurvesHoudini");
			if(m_spSurfaceI.isValid())
			{
				sp<SurfaceCurvesHoudini> spSurface=m_spSurfaceI;
				if(spSurface.isValid())
				{
					spSurface->setGdp(m_pGdp);
					spSurface->setGroup(a_group);
					spSurface->setGrain(subCount);
					spSurface->setOpenCurve(TRUE);
				}
			}
		}
		else
		{
			//* all triangles (presumed)
			m_spSurfaceI=registry()->create("SurfaceI.SurfaceTrianglesHoudini");
			if(m_spSurfaceI.isValid())
			{
				sp<SurfaceTrianglesHoudini> spSurface=m_spSurfaceI;
				if(spSurface.isValid())
				{
					spSurface->setGdp(m_pGdp);
					spSurface->setGroup(a_group);
				}
			}
		}
	}

	safeUnlock();

	return m_spSurfaceI;
}

sp<SurfaceI> SurfaceAccessibleHoudini::subSurface(U32 a_subIndex,String a_group)
{
//	feLog("SurfaceAccessibleHoudini::subSurface %d \"%s\"\n",
//			a_subIndex,a_group.c_str());

	if(!m_pGdp)
	{
		return sp<SurfaceI>(NULL);
	}

	safeLock();

	if(m_primitiveMap.empty())
	{
		initPrimitiveMap();
	}

	const I32 lookupIndex=lookupPrimitive(a_subIndex);
	if(lookupIndex< -1)
	{
		safeUnlock();

		return sp<SurfaceI>(NULL);
	}

#ifdef FE_HOUDINI_USE_GA
	const I32 primitiveCount=m_pGdp->getNumPrimitives();
#else
	const I32 primitiveCount=m_pGdp->primitives().entries();
#endif

	const I32 primitiveIndex=(lookupIndex>=0)? lookupIndex: a_subIndex;
	if(primitiveIndex>=primitiveCount)
	{
		safeUnlock();

		return sp<SurfaceI>(NULL);
	}


#ifdef FE_HOUDINI_USE_GA
	const GEO_Primitive* pPrimitive=m_pGdp->getGEOPrimitive(
			m_pGdp->primitiveOffset(primitiveIndex));
#else
	const GEO_Primitive* pPrimitive=m_pGdp->primitives()(primitiveIndex);
#endif

	if(!pPrimitive)
	{
		safeUnlock();

		return sp<SurfaceI>(NULL);
	}

	if(lookupIndex>=0)
	{

#ifdef FE_HOUDINI_USE_GA
		if(pPrimitive->getTypeId().get()==GEO_PRIMSPHERE)
#else
		if(pPrimitive->getPrimitiveId()==GEOPRIMSPHERE)
#endif
		{
			sp<SurfaceSphere> spSphere=
					registry()->create("SurfaceI.SurfaceSphere");
			if(!spSphere.isValid())
			{
				safeUnlock();

				return sp<SurfaceI>(NULL);
			}

			const GEO_PrimSphere* pPrimSphere=
					dynamic_cast<const GEO_PrimSphere*>(pPrimitive);

			UT_Matrix4 utMatrix4;
			pPrimSphere->getTransform4(utMatrix4);
			SpatialTransform matrix;
			set(matrix,utMatrix4.data());

//			fpointOffseteLog("SurfaceAccessibleHoudini::subSurface(%d) prim %d"
//					" sphere matrix\n%s\n",
//					a_subIndex,lookupIndex,c_print(matrix));

			spSphere->setCenter(matrix.translation());
			spSphere->setRadius(matrix(0,0));

			safeUnlock();

			return spSphere;
		}

#ifdef FE_HOUDINI_USE_GA
		if(pPrimitive->getTypeId().get()==GEO_PRIMCIRCLE)
#else
		if(pPrimitive->getPrimitiveId()==GEOPRIMCIRCLE)
#endif
		{
			sp<SurfaceDisk> spDisk=
					registry()->create("SurfaceI.SurfaceDisk");
			if(!spDisk.isValid())
			{
				safeUnlock();

				return sp<SurfaceI>(NULL);
			}

			const GEO_PrimCircle* pPrimCircle=
					dynamic_cast<const GEO_PrimCircle*>(pPrimitive);

			UT_Matrix4 utMatrix4;
			pPrimCircle->getTransform4(utMatrix4);
			SpatialTransform matrix;
			set(matrix,utMatrix4.data());

			UT_Vector3 utVector3=pPrimCircle->computeNormal();

//			feLog("SurfaceAccessibleHoudini::subSurface(%d) prim %d"
//					" disk matrix\n%s\n",
//					a_subIndex,lookupIndex,c_print(matrix));

			spDisk->setCenter(matrix.translation());
			spDisk->setRadius(matrix(0,0));
			spDisk->setSpan(
					SpatialVector(utVector3[0],utVector3[1],utVector3[2]));

			safeUnlock();

			return spDisk;
		}

#ifdef FE_HOUDINI_USE_GA
		if(pPrimitive->getTypeId().get()==GEO_PRIMTUBE)
#else
		if(pPrimitive->getPrimitiveId()==GEOPRIMTUBE)
#endif
		{
			sp<SurfaceCylinder> spCylinder=
					registry()->create("SurfaceI.SurfaceCylinder");
			if(!spCylinder.isValid())
			{
				safeUnlock();

				return sp<SurfaceI>(NULL);
			}

			const GEO_PrimTube* pPrimTube=
					dynamic_cast<const GEO_PrimTube*>(pPrimitive);

			UT_Matrix4 utMatrix4;
			pPrimTube->getTransform4(utMatrix4);
			SpatialTransform matrix;
			set(matrix,utMatrix4.data());

			const Real taper=pPrimTube->getTaper();

//			feLog("SurfaceAccessibleHoudini::subSurface(%d) prim %d"
//					" cylinder taper %.6G matrix\n%s\n",
//					a_subIndex,lookupIndex,taper,c_print(matrix));

			const Real scaleX=magnitude(matrix.column(0)); // same as scaleZ

//			feLog("scale %.6G %.6G %.6G\n",scaleX,scaleY,scaleZ);

		const SpatialVector span=matrix.column(1);	//* already scaled

			spCylinder->setLocation(matrix.translation()-0.5*span);
			spCylinder->setSpan(span);
			spCylinder->setBaseRadius(scaleX);
			spCylinder->setEndRadius(scaleX*taper);

			safeUnlock();

			return spCylinder;
		}
	}

	//* HACK presuming they're all the same
	const GEO_Face* pGeoFace=dynamic_cast<const GEO_Face*>(pPrimitive);
	const BWORD openPoly=pGeoFace && !pGeoFace->isClosed();
	if(openPoly)
	{
		//* HACK presuming uniform CV counts
		const U32 subCount=pGeoFace->getVertexCount();

		sp<SurfaceCurvesHoudini> spCurves=
				registry()->create("SurfaceI.SurfaceCurvesHoudini");
		if(!spCurves.isValid())
		{
			safeUnlock();

			return sp<SurfaceI>(NULL);
		}

		spCurves->setGdp(m_pGdp);
		spCurves->setGroup(a_group);
		spCurves->setGrain(subCount);
		spCurves->setOpenCurve(TRUE);
		if(lookupIndex>=0)
		{
			spCurves->setSubIndex(a_subIndex);
		}

		safeUnlock();

		return spCurves;
	}

	sp<SurfaceTrianglesHoudini> spTriangles=
			registry()->create("SurfaceI.SurfaceTrianglesHoudini");
	if(!spTriangles.isValid())
	{
		safeUnlock();

		return sp<SurfaceI>(NULL);
	}

	spTriangles->setGdp(m_pGdp);
	spTriangles->setGroup(a_group);
	if(lookupIndex>=0)
	{
		spTriangles->setSubIndex(a_subIndex);
	}

	safeUnlock();

	return spTriangles;
}

} // namespace
