import os
import sys
import shutil
forge = sys.modules["forge"]

# NOTE: The Houdini version of boost may be different.
# The boost-regex appears to be sensitive and they don't provide the dso.
# Moving aside their include/boost seems to fix regex,
# but it leaves their other boost dso's cross-versioned with system includes.

# NOTE: Houdini ode includes also conflict
# Try moving aside their include/ode.

def setup(module):
	srcList = [	"houdini.pmh",
				"HoudiniBrush",
				"HoudiniDraw",
				"HoudiniSOP",
				"SurfaceAccessibleHoudini",
				"SurfaceAccessorHoudini",
				"SurfaceCurvesHoudini",
				"SurfaceTrianglesHoudini",
				"assertHoudini",
				"houdiniDL" ]

	# force all of FE to use Houdini's version of boost
#	forge.includemap['houdini'] = os.path.join(os.getenv('HT'),'include')

	dll = module.DLL( "fexHoudiniDL", srcList )

	# this overrides previous -W flags in codegen
	module.cppmap = {}
	module.cppmap['houdini'] = ' -Wno-deprecated -Wall -W -Wno-parentheses -Wno-sign-compare -Wno-reorder -Wno-uninitialized -Wunused -Wno-unused-parameter'

	deplibs = forge.corelibs+ [
				"fexDrawDLLib",
				"fexMetaDLLib",
				"fexSurfaceDLLib",
				"fexWindowLib"	]

	dll.linkmap = {}
	if forge.fe_os == "FE_LINUX":
		# needed for gcc 4.1, not for gcc 4.4, but not sure otherwise
		if forge.compiler_version <= 4.1:
			module.cppmap['houdini'] += ' -DNEED_SPECIALIZATION_STORAGE'

		dll.linkmap["houdinilibs"] = "-shared"
		dll.linkmap["houdinilibs"] += " -L"+os.getenv('HDSO')
		dll.linkmap["houdinilibs"] += " -lHoudiniUI -lHoudiniOPZ -lHoudiniOP3 -lHoudiniOP2 -lHoudiniOP1 -lHoudiniSIM -lHoudiniGEO -lHoudiniPRM -lHoudiniUT -lHoudiniAPPS3"
		dll.linkmap["houdinilibs"] += " -Xlinker -rpath='$ORIGIN/../../'"
	elif forge.fe_os == "FE_WIN32" or forge.fe_os == "FE_WIN64":
		dll.linkmap["houdinilibs"] = "TODO"

	forge.deps( ["fexHoudiniDLLib"], deplibs )

	houdini_major = 12
	houdini_major_string = os.getenv('HOUDINI_MAJOR_RELEASE')
	if houdini_major_string:
		houdini_major = int(houdini_major_string)
	module.cppmap['houdini'] += ' -DFE_HOUDINI_MAJOR=' + str(houdini_major)

	houdini_path = os.getenv('HT')
	if houdini_path:
		module.includemap = { 'houdini' : os.path.join(houdini_path,'include') }

	# make sym links for houdini to look at
	# TODO should work on non-linux
	houdiniDso = 'libfexHoudiniDL.so'
	houdiniPath = os.path.join(forge.libPath, 'houdini')
	houdiniDsoPath = os.path.join(houdiniPath, 'dso')
	houdiniIconPath = os.path.join(houdiniPath, 'config', 'Icons')
	houdiniToolbarPath = os.path.join(houdiniPath, 'toolbar')
	if os.path.isdir(houdiniDsoPath) == 0:
		os.makedirs(houdiniDsoPath, 0755)
	if os.path.isdir(houdiniIconPath) == 0:
		os.makedirs(houdiniIconPath, 0755)
	if os.path.isdir(houdiniToolbarPath) == 0:
		os.makedirs(houdiniToolbarPath, 0755)

	source = os.path.join('..', '..', houdiniDso)
	dest = os.path.join(houdiniDsoPath, houdiniDso)
	if os.path.lexists(dest) == 0:
		os.symlink(source, dest)

	for icon in ['FE.svg', 'FE_alpha.svg', 'FE_beta.svg']:
		source = os.path.join('..', '..', '..', '..', '..', 'doc' , 'image', icon)
		dest = os.path.join(houdiniIconPath, icon)
		if os.path.lexists(dest) == 0:
			os.symlink(source, dest)

	shelf = os.path.join(houdiniToolbarPath, 'FreeElectron.shelf')
	shutil.copyfile('ext/houdini/FreeElectron.shelf', shelf)

	fe_houdini_libs = 'fexHoudiniDL'
	if os.path.exists('lab/grass'):
		fe_houdini_libs += ':fexGrassDL'
		sys.stdout.write("grass ")
	if os.path.exists('lab/vegetation'):
		fe_houdini_libs += ':fexVegetationDL'
		sys.stdout.write("vegetation ")
	module.cppmap['houdini'] += ' -DFE_HOUDINI_LIBS=' + fe_houdini_libs

	test = module.Module('test')
	test.cppmap = module.cppmap

def auto(module):
	test_file = """
#include <SOP/SOP_API.h>

int main(void)
{
	return(0);
}
    \n"""

	houdini_path = os.getenv('HT')
	if houdini_path:
		forge.includemap['houdini'] = os.path.join(houdini_path,'include')

	return forge.cctest(test_file)
