/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __surface_SurfaceRange_h__
#define __surface_SurfaceRange_h__

namespace fe
{

/**************************************************************************//**
	@brief SurfaceRange of jobs

	@ingroup surface
*//***************************************************************************/
class FE_DL_EXPORT SurfaceRange: public Counted
{
	public:

	class Span: public Vector<2,Real>
	{
		public:
					Span(void):
						Vector<2,Real>(0.0,-1.0)							{}
	};

	class MultiSpan
	{
		public:
								MultiSpan(void)								{}

			void				clear(void)
								{	m_spanArray.clear(); }
			U32					spanCount(void) const
								{	return m_spanArray.size(); }
			Span&				span(U32 a_index)
								{
									FEASSERT(a_index<m_spanArray.size());
									return m_spanArray[a_index];
								}
	const	Span&				span(U32 a_index) const
								{
									FEASSERT(a_index<m_spanArray.size());
									return m_spanArray[a_index];
								}

			void				add(Real a_start,Real a_end);
			void				add(const MultiSpan& rOther);

			U32					valueCount(void) const;

			String				brief(void) const;
			String				dump(void) const;

			std::vector<Span>	m_spanArray;
	};

	class Iterator
	{
		public:
								Iterator(sp<SurfaceRange> a_spSurfaceRange):
									m_spSurfaceRange(a_spSurfaceRange),
									m_atomicIndex(0),
									m_spanIndex(-1),
									m_valueCount(0),
									m_valueIndex(0),
									m_value(0.0)
								{
									FEASSERT(m_spSurfaceRange.isValid());
									m_atomicCount=
											m_spSurfaceRange->atomicCount();
									advance();
								}

			Real				value(void)		{ return m_value; }
			BWORD				atEnd(void)
								{	return m_atomicIndex>m_atomicCount+1; }

			void				step(void)
								{
									m_valueIndex++;
									if(m_valueIndex>=m_valueCount)
									{
										advance();
									}
									else
									{
										m_value+=1.0;
									}
								}

		private:

			void				advance(void);

			sp<SurfaceRange>	m_spSurfaceRange;
			U32					m_atomicCount;
			U32					m_atomicIndex;
			I32					m_spanIndex;
			U32					m_valueCount;
			U32					m_valueIndex;
			Real				m_value;
	};

					SurfaceRange(void)										{}

		void		clear(void)
					{
						m_atomArray.clear();
						m_postAtomic.clear();
						m_nonAtomic.clear();
					}

		MultiSpan&	addAtomic(void);

		U32			atomicCount(void) const	{ return m_atomArray.size(); }

		MultiSpan&	atomic(U32 a_index)
					{	return m_atomArray[a_index]; }
const	MultiSpan&	atomic(U32 a_index) const
					{	return m_atomArray[a_index]; }

		MultiSpan&	postAtomic(void)		{ return m_postAtomic; }
		MultiSpan&	nonAtomic(void)			{ return m_nonAtomic; }

		BWORD		empty(void) const
					{	return !m_atomArray.size() &&
								!m_postAtomic.spanCount() &&
								!m_nonAtomic.spanCount(); }

		Iterator	begin(void)	{ return Iterator(sp<SurfaceRange>(this)); }

					/** @brief merge atoms into counts of the given granularity

						The granularity size is only a goal.
						The results can be various sizes
						above and below the goal.
						No atoms are broken up to make them smaller.
					*/
		sp<SurfaceRange> combineAtoms(I32 a_size) const;

		String		brief(void) const;
		String		dump(void) const;

	private:

		std::vector<MultiSpan>	m_atomArray;	//* each thread-safe piece
		MultiSpan				m_postAtomic;	//* unsafe to do unthreaded
		MultiSpan				m_nonAtomic;	//* always safe
};

inline String print(const sp<SurfaceRange>& a_rspSurfaceRange)
{
	return a_rspSurfaceRange.isNull()? "NULL": a_rspSurfaceRange->brief();
}

} // namespace

#endif // __surface_SurfaceRange_h__
